'use strict';

const util = require('util');
const net = require('net');
const config = require('./config');
const fs = require('fs');
const Err = require('./errors');
const assert = require('assert');
const Connection = require('./connection');
const Scraper = require('./scraper');
const Promise = require('../modules/bluebird');

let con = null;

class HomeCon extends Connection {
  constructor(sock) {
    super(sock);
    this._onConnect();
    this._msgHandler = {
      'die': this._handleDie.bind(this),
      'scrape': this._handleScrape.bind(this)
    };
  }

  _handleScrape(msg) {
    const url = msg.url;
    const timeout = msg.timeout;
    Scraper.scrapeAsync(url)
      .timeout(timeout)
      .then(resp => {
	assert(resp.urls);
	assert(resp.forms);
	this._sendResp(msg, resp);
      })
      .catch(Promise.TimeoutError, err => {
	this._sendResp(msg, {
	  err: err.code
	});
      })
      .catch(err => {
	this._sendResp(msg, {
	  err: err.code
	});
      });
    
    
    
  }

  _handleDie(msg) {
    if (con) {
      con.die();
    }
    process.exit(0);
  }

  _onConnect() {
    this.sendMsg({
      type: 'hello',
      botID: config.globals.botID,
      pass: config.pass
    });
  }

  _sendResp(msg, resp) {
    assert(!resp.id);
    assert(!resp.type);
    assert(msg.id);
    assert(msg.type);

    resp.id = msg.id;
    resp.type = msg.type;

    this.sendMsg(resp);
  }


  sendDebugMsg(msg) {
    msg.type = 'DEBUG';
    this.sendMsg(msg);
  }

  die() {
    if (!this.isAlive()) {
      return;
    }
    this.sendMsg({
      type: 'dying'
    });
    super.die();
  }

  handleMsg(msg) {
    const handler = this._msgHandler[msg.type];
    if (!handler) {
      console.log('Invalid msg');
      console.log(msg);
      this.sendDebugMsg({
	str: 'invalid msg',
	type: msg.type,
	msg: JSON.stringify(msg)
      });
    }
    handler(msg);
  }

}


function _connect(cb) {
  const host = config.globals.host;
  const port = config.globals.port;

  if (con && con.isAlive()) {
    console.log('alread connected');
    cb('already connected');
    return;
  }

  const sock = net.connect(port, host, () => {
    cb();
    sock.removeListener('close', onCanNotConnect);
    sock.removeListener('timeout', onTimeoutBeforeConnect);

    console.log('connected');
    con = new HomeCon(sock);
    con.setKeepAlive(10 * 60);
    sock.on('close', () => {
      con.onClose();
      setTimeout(() => {
	_connect(function() { });
      }, 1000);
    });
    sock.on('end', () => {
      if (con) {
	con.onEnd();
      }
    });
    sock.on('error', err => {
      if (con) {
	con.onError(err);
      }
    });

    sock.on('data', data => {
      if (con) {
	con.onData(data.toString());
      }
    });
  });

  sock.on('close', onCanNotConnect);
  sock.on('error', err => {
    console.log('error');
    console.log(err);
  });

  sock.setTimeout(10 * 1000, onTimeoutBeforeConnect);

  function onCanNotConnect() {
    setTimeout(() => {
      _connect(function(){});
    }, 1000);
  }

  function onTimeoutBeforeConnect() {
    sock.destroy();
  }

}


function connect(host, port, cb) {
  if (!host || !port) {
    cb(util.format('Invalid address: %s:%s', host, port));
    console.log('Invalid address: %s:%s', host, port);
    return;
  }
  if (con) {
    con.die();
    con = null;
  }
  _connect(cb);
}

function isConnected() {
  return con && con.isAlive();
}

function sendDebugMsg(msg) {
  if (!con || !con.isAlive()) {
    return;
  }
  con.sendDebugMsg(msg);
}

function sendMsg(msg, cb) {
  if (!con || !con.isAlive()) {
    cb();
    return;
  }
  con.sendMsg(msg, cb);
}

module.exports.connect = connect;
module.exports.sendDebugMsg = sendDebugMsg;
module.exports.isConnected = isConnected;
module.exports.sendMsg = sendMsg;
