'use strict';

const net = require('net');
const Utils = require('./utils');
const Promise = require('../modules/bluebird');
const config = require('./config');


const instances = [];

class ReverseProxyFront {

  constructor(bindIP, peerIP) {
    this._bindIP = bindIP;
    this._peerIP = peerIP;
    this._server = net.createServer(this._onConnection.bind(this));
    this._server.on('error', err => {
      if (this._onListenError) {
        this._onListenError(err);
      }
      this.delete();
    });

    this._clientCons = [];
    this._peerCons = [];

    this._intervalID = setInterval(() => {
      console.log('%d client cons, %d peer cons', this._clientCons.length,
		 this._peerCons.length);
    }, 3000);

    Promise.promisifyAll(this, {
      filter: function() {
	return true;
      }
    });

    instances.push(this);
  }

  get bindIP() {
    return this._bindIP;
  }

  get peerIP() {
    return this._peerIP;
  }

  get bindPort() {
    return this._bindPort;
  }

  _onPeerCon(sock) {
    const item = {
      sock
    };
    this._peerCons.push(item);

    sock.on('close', () => {
      const index = this._peerCons.indexOf(item);
      if (index === -1) {
        return;
      }
      this._peerCons.splice(index, 1);
    });
  }

  _onClientCon(sock) {
    const item = {
      sock
    };
    this._clientCons.push(item);

    sock.on('close', () => {
      const index = this._clientCons.indexOf(item);
      if (index === -1) {
        return;
      }
      this._clientCons.splice(index, 1);
    });
  }

  _makePeers() {
    if (this._clientCons.length === 0 || this._peerCons.length === 0) {
      return;
    }
    const clientCon = this._clientCons.shift();
    const peerCon = this._peerCons.shift();

    const clientSock = clientCon.sock;
    const peerSock = peerCon.sock;

    clientSock.on('data', data => {
      peerSock.write(data);
    });
    clientSock.on('error', err => {
      peerSock.end();
      peerSock.destroy();
    });
    clientSock.on('close', () => {
      peerSock.end();
    });
    clientSock.on('end', () => {
      peerSock.end();
    });
    peerSock.on('data', data => {
      clientSock.write(data);
    });
    peerSock.on('error', err => {
      clientSock.end();
      clientSock.destroy();
    });
    peerSock.on('close', () => {
      clientSock.end();
    });
    peerSock.on('end', () => {
      clientSock.end();
    });

  }

  _onConnection(sock) {
    console.log('connection received from: %s', sock.remoteAddress);
    sock.on('error', err => { });
    if (!sock.remoteAddress) {
      return;
    }
    if (sock.remoteAddress === this._peerIP) {
      console.log('peerconnection');
      this._onPeerCon(sock);
    } else {
      console.log('client connection');
      this._onClientCon(sock);
    }
    this._makePeers();
  }

  listen(cb) {
    if (this._bindPort) {
      process.nextTick(() => {
	cb(null, this._bindPort);
      });
      return;
    }
    this._listen(this._bindIP, 10, (err, port) => {
      if (err) {
        cb(err);
        return;
      }
      this._bindPort = port;
      cb(null, port);
    });
  }

  delete() {
    this._server.close();
    for (const item of this._clientCons) {
      const sock = item.sock;
      sock.end();
    }
    for (const item of this._peerCons) {
      const sock = item.sock;
      sock.end();
    }
    clearInterval(this._intervalID);
    const index = instances.indexOf(this);
    if (index === -1) {
      return;
    }
    instances.splice(index, 1);
  }

  _listen(bindIP, nTries, cb) {
    const port = Utils.getRandomInt(config.portLow, config.portHigh);
    this._server.listen(port, bindIP, () => {
      cb(null, port);
      cb = Function.prototype;
    });
    this._onListenError = err => {
      if (nTries > 1) {
        this._listen(bindIP, nTries - 1, cb);
        return;
      }
      cb(err);
      cb = Function.prototype;
    };

  }

  static toArray() {
    return instances.slice();
  }

  static find(bindIP, peerIP) {
    for (const item of instances) {
      if (item.bindIP !== bindIP) {
	continue;
      }
      if (item.peerIP !== peerIP) {
	continue;
      }
      if (!item.bindPort) {
	continue;
      }
      return item;
    }
    return null;
  }
}

module.exports = ReverseProxyFront;
