'use strict';

const net = require('net');
const assert = require('assert');

const STATES = {
  new: 1,
  identifierReceived: 2,
  successfull: 3,
  failed: 4
};

class Socks5 {
  constructor(clientSock, initialRecData, localIP, IPPairs) {
    assert(localIP);
    if (!IPPairs) {
      IPPairs = new Map();
    }
    this._IPPairs = IPPairs;
    this._localIP = localIP;
    this._clientSock = clientSock;
    this._destSock = null;
    this._destBuf = new Buffer([]);

    console.log('changing state from %d to %d',
		0, STATES.new);
    this._state = STATES.new;
    this._clientRecData = new Buffer([]);
    this._isFresh = true;
    this._onBusy = function() {};

    this._clientSock.on('data', this._onData.bind(this));
    this._clientSock.on('error', this._onError.bind(this));
    this._clientSock.on('end', this._onEnd.bind(this));
    this._clientSock.on('close', this._onClose.bind(this));

    if (initialRecData) {
      this._onData(initialRecData);
    }
  }

  _onDataWhenNew() {
    console.log('on data when new');
    if (this._clientRecData.length < 2) {
      return;                   // not enough data
    }

    const version = this._clientRecData[0];
    if (version !== 5) {
      console.log('version not 5: %s', version);
      // todo: close connection
      return;
    }

    const nMethods = this._clientRecData[1];
    const length = 2 + nMethods;
    if (this._clientRecData.length < length) {
      return;                   // not enough data
    }

    this._clientRecData = this._clientRecData.slice(length);
    const sndbuf = new Buffer([5, 0]); // socks version 5; no auth required
    this._clientSock.write(sndbuf);
    console.log('changing state from %d to %d',
		this._state, STATES.identifierReceived);
    this._state = STATES.identifierReceived;
  }

/*
 * currentExternalIP: ip where client is connected
 */
  _chooseLocalIP(defaultLocalIP, currentExternalIP) {
    assert(defaultLocalIP);
    if (defaultLocalIP !== '0.0.0.0') {
      return defaultLocalIP;
    }
    assert(currentExternalIP);
    const result = this._IPPairs.get(currentExternalIP);
    assert(result);
    if (!result) {
      // todo: wtf? debug
      return defaultLocalIP;
    }
    return result;
  }

  _onDataWhenIdentifierReceived() {
    const data = this._clientRecData;
    if (data.length < 6) {
      return;                   // not enough dat
    }
    const version = data[0];
    if (version !== 5) {
      // invalid version
      // todo: close connection;
      return;
    }

    const cmd = data[1];
    if (cmd !== 1) {
      if (cmd === 2 || cmd === 3) {
        console.log('not implemented..');
        // todo: not implemented
      } else {
        console.log('invalid data'); // todo: do something about it
      }
      return;
    }

    let msgLen = 6;
    const atyp = data[3];
    let dstIp = null;
    let dstPort = null;
    if (atyp === 1) {           // ipv5
      if (data.length < 10) {
        return;                 // not enough data
      }
      console.log('extract ip and port from here');
      dstIp = data[4] + '.' + data[5] + '.' + data[6] + '.' + data[7];
      dstPort = data[8];
      dstPort <<= 8;
      dstPort += data[9];
      console.log(data);
      msgLen += 4;
    } else if (atyp === 3) {
      const nameLen = data[4];
      if (data.length < nameLen + 7) {
        return;                 // not enough data
      }
      dstIp = data.slice(5, 5 + nameLen);
      dstIp = dstIp.toString();
      dstPort = data[5 + nameLen];
      dstPort <<= 8;
      dstPort += data[6 + nameLen];
      msgLen += nameLen + 1;
    } else if (atyp === 6) {
      dstIp = data.slice(4, 4 + 16);
      dstPort = data[20];
      dstPort <<= 8;
      dstPort += data[21];
      msgLen += 16;
    }
    this._clientRecData = this._clientRecData.slice(msgLen);

    console.log("connecting to: %s:%d", dstIp, dstPort);
    console.log('changing state from %d to %d',
		this._state, STATES.successfull);
    this._state = STATES.successfull;
    if (dstIp && dstPort) {
      this._destSock = net.connect({
        port: dstPort,
        host: dstIp,
//	localAddress: this._localIP
	localAddress: this._chooseLocalIP(this._localIP,
					  this._clientSock.localAddress)
      }, () => {
        this._destSock.removeListener('error', onErr);
        // TODO: this hack might speed things up considerebly (?). execute the following lines immediately, not in the callback, as it is now.
        var sndbuf = new Buffer([5, 0, 0, 1, 0, 0, 0, 0, 0, 0]);
        this._clientSock.write(sndbuf);
        this._destSock.removeListener('error', onErr);
        this._destSock.on('data', data => {
          this._clientSock.write(data);
        });
        this._destSock.on('error', err => {
          console.log("destsock err");
          console.log(err);
          this._clientSock.end();
          this._clientSock.destroy();
        });
        this._destSock.on('end', () => {
	  console.log('dest socket ended');
          this._clientSock.end();
        });
        this._destSock.on('close', () => {
	  console.log('dest socket closed');	  
          this._clientSock.end();
        });
      });
      this._destSock.on('error', onErr);

      const socks5 = this;
      function onErr() {
        socks5._clientSock.write('error');
        socks5._clientSock.end();
        socks5._clientSock.destroy();
        socks5._state = STATES.failed;
        console.log('could not connect to dst server');
      }
    }
  }

  _onDataWhenSuccessfull(data) {
    if (this._destSock === null) {
      // not yet connected. save somewhere
      this._clientRecData = Buffer.concat([this._clientRecData, data]);
      return;
    }
    if (this._clientRecData.length > 0) {
      const oldData = this._clientRecData;
      this._destSock.write(oldData);
      this._clientRecData = new Buffer([]);
    }
    this._destSock.write(data);
  }

  _onDataWhenFailed() {
    this._clientSock.end();
    this._clientSock.destroy();
    // todo
  }

  _onData(data) {
    console.log('ondata');
    console.log('state is: %s', this._state);
    console.log(data);
    console.log(data.toString());    
    if (this._isFresh) {
      this._onBusy();
      this._isFresh = false;
    }

    if (this._state === STATES.new) {
      this._clientRecData = Buffer.concat([this._clientRecData, data]);
      this._onDataWhenNew();
      return;
    }
    if (this._state === STATES.identifierReceived) {
      this._clientRecData = Buffer.concat([this._clientRecData, data]);
      this._onDataWhenIdentifierReceived();
      return;
    }
    if (this._state === STATES.successfull) {
      this._onDataWhenSuccessfull(data);
      return;
    }
    if (this._state === STATES.failed) {
      this._onDataWhenFailed(data);
      return;
    }
  }

  _onEnd() {
    console.log('client socket ended'); 
    if (this._isFresh) {
      this._onBusy();
      this._isFresh = false;
    }
  }

  _onError(err) {
    if (this._isFresh) {
      this._onBusy();
      this._isFresh = false;
    }
  }

  _onClose() {
    console.log('client socket closed'); 
    if (this._isFresh) {
      this._onBusy();
      this._isFresh = false;
    }
  }

  onErr() {
    if (this._isFresh) {
      this._onBusy();
      this._isFresh = false;
    }
  }

  setOnBusy(cb) {
    if (cb) {
      this._onBusy = cb;
      return;
    }
    this._onBusy = function() {};
  }
};


module.exports = Socks5;
