'use strict';

const assert = require('assert');
const Err = require('./errors');
const URL = require('url');
const os = require('os');
const net = require('net');
const http = require('http');
const Promise = require('../modules/bluebird');

class Utils {

  static getRandomInt(min, max) {
    return Math.floor(Math.random() * (max - min + 1)) + min;
  };

  static _isInternalIP(ip) {
    const arr = ip.split('.');
    if (arr[0] === '10') {
      return true;
    }
    if (arr[0] === '172' && arr[1] === '16') {
      return true;
    }
    if (arr[0] === '192' && arr[1] === '168') {
      return true;
    }
    return false;
  }

  static getIPS() {
    let interfaces;
    try {
      interfaces = os.networkInterfaces();
    } catch (e) {
      console.log(e);
      return [];
    }
    const addresses = [];
    for (let k in interfaces) {
      if (!interfaces.hasOwnProperty(k)) {
        continue;
      }
      for (const address of interfaces[k]) {
        const ip = address.address;
        if (address.family !== 'IPv4' || address.internal) {
          continue;
        }
        if (addresses.indexOf(ip) !== -1) {
          continue;
        }
        addresses.push(ip);
      }
    }
    return addresses;
  };


  static canConnect(ip, port, cb) {
    const sock = net.connect(port, ip, () => {
      cb(null, true);
      cb = Function.prototype;
      sock.end();
      sock.destroy();
    });
    sock.setTimeout(10 * 1000);
    sock.on('timeout', () => {
      sock.end();
      sock.destroy();
    });
    sock.on('error', err => {
      // todo: check error type
    });
    sock.on('close', () => {
      cb(null, false);
      cb = Function.prototype;
    });
  }

  static canConnectFrom(bindIP, ip, port, cb) {
    const options = {
      port: port,
      host: ip,
      localAddress: bindIP
    };
    const sock = net.connect(options, () => {
      cb(null, true);
      cb = Function.prototype;
      sock.end();
      sock.destroy();
    });
    sock.setTimeout(60 * 1000);
    sock.on('timeout', () => {
      sock.end();
      sock.destroy();
    });
    sock.on('error', err => {
      // todo: check error type
    });
    sock.on('close', () => {
      cb(null, false);
      cb = Function.prototype;
    });
  }


  static findIPInfo(ip, cb) {
    //const url = 'http://ipinfo.io/' + ip;
    const url = 'http://ipinfo.io/';
    const parsed = URL.parse(url);
    const options = {
      hostname: parsed.hostname,
      path: parsed.path,
      localAddress: ip
    };

    try {
      const request = http.get(url, res => {
	let data = '';
	res.on('data', d => {
	  data += d.toString();
	});
	res.on('end', () => {
	  cb(null, data);
	});
      });

      request.setTimeout(10 * 1000, function() {
	request.abort();
	cb(null, '{}');
      });

      request.on('error', err => {
	cb(null, '{}');
      });
    } catch (e) {
      cb(null, '{}');
    }

  }

  static findExternalIP(ip, cb) {
    console.log(ip);
//    ip = '123.23.23.23';
    const url = 'http://api.ipify.org?format=json';
    const parsed = URL.parse(url);
    const options = {
      hostname: parsed.hostname,
      path: parsed.path,
      localAddress: ip
    };
    const req = http.get(options, resp => {
      let data = '';
      resp.on('data', d => {
	data += d.toString();
      });
      resp.on('end', () => {
	console.log(data);
	try {
	  const parsed = JSON.parse(data);
	  assert(parsed.ip);
	  cb(null, parsed.ip);
	} catch (err) {
	  const error = new Err.SysError(err);
	  error.setCustomMsg(data);
	  cb(error);
	}
      });
    });
    req.on('error', err => {
      console.log(err);
      console.log(err.toString());
      cb(new Err.SysError(err));
    });
  }

};

Promise.promisifyAll(Utils, {
  filter: function() {
    return true;
  }
});

module.exports = Utils;
